@@ -62,6 +62,10 @@ namespace {
     OBVIOUS_KEYWORD,    // The OBVIOUS keyword.
     OMITTED_KEYWORD,    // The OMITTED keyword.
     QED_KEYWORD,        // The QED keyword.
+    WEAK_FAIRNESS,      // The WF_ keyword.
+    STRONG_FAIRNESS,    // The SF_ keyword.
+    PCAL_START,         // Notifies scanner of start of PlusCal block.
+    PCAL_END,           // Notifies scanner of end of PlusCal block.
     ERROR_SENTINEL      // Only valid if in error recovery mode.
   };
 
@@ -316,6 +320,8 @@ namespace {
     Lexeme_VARIABLES_KEYWORD,
     Lexeme_PROOF_STEP_ID,
     Lexeme_IDENTIFIER,
+    Lexeme_WEAK_FAIRNESS,
+    Lexeme_STRONG_FAIRNESS,
     Lexeme_OTHER,
     Lexeme_END_OF_FILE
   };
@@ -339,6 +345,7 @@ namespace {
     LexState_R_SQUARE_BRACKET,
     LexState_R_CURLY_BRACE,
     LexState_R_ANGLE_BRACKET,
+    LexState_S, LexState_SF_,
     LexState_RIGHT_ARROW,
     LexState_COMMENT_START,
     LexState_BLOCK_COMMENT_START,
@@ -355,6 +362,7 @@ namespace {
     LexState_Q, LexState_QED,
     LexState_T, LexState_THE, LexState_THEN, LexState_THEOREM,
     LexState_V, LexState_VARIABLE, LexState_VARIABLES,
+    LexState_W, LexState_WF_,
     LexState_IDENTIFIER,
     LexState_PROOF_LEVEL_NUMBER,
     LexState_PROOF_LEVEL_STAR,
@@ -412,8 +420,10 @@ namespace {
         if ('O' == lookahead) ADVANCE(LexState_O);
         if ('P' == lookahead) ADVANCE(LexState_P);
         if ('Q' == lookahead) ADVANCE(LexState_Q);
+        if ('S' == lookahead) ADVANCE(LexState_S);
         if ('T' == lookahead) ADVANCE(LexState_T);
         if ('V' == lookahead) ADVANCE(LexState_V);
+        if ('W' == lookahead) ADVANCE(LexState_W);
         if (L'∧' == lookahead) ADVANCE(LexState_LAND);
         if (L'∨' == lookahead) ADVANCE(LexState_LOR);
         if (L'〉' == lookahead) ADVANCE(LexState_R_ANGLE_BRACKET);
@@ -654,6 +664,13 @@ namespace {
         ACCEPT_LEXEME(Lexeme_QED_KEYWORD);
         if (is_identifier_char(lookahead)) ADVANCE(LexState_IDENTIFIER);
         END_LEX_STATE();
+      case LexState_S:
+        ACCEPT_LEXEME(Lexeme_IDENTIFIER);
+        if (is_next_codepoint_sequence(lexer, "F_")) ADVANCE(LexState_SF_);
+        END_LEX_STATE();
+      case LexState_SF_:
+        ACCEPT_LEXEME(Lexeme_STRONG_FAIRNESS);
+        END_LEX_STATE();
       case LexState_T:
         ACCEPT_LEXEME(Lexeme_IDENTIFIER);
         if (is_next_codepoint_sequence(lexer, "HE")) ADVANCE(LexState_THE);
@@ -684,6 +701,13 @@ namespace {
         ACCEPT_LEXEME(Lexeme_VARIABLES_KEYWORD);
         if (is_identifier_char(lookahead)) ADVANCE(LexState_IDENTIFIER);
         END_LEX_STATE();
+      case LexState_W:
+        ACCEPT_LEXEME(Lexeme_IDENTIFIER);
+        if (is_next_codepoint_sequence(lexer, "F_")) ADVANCE(LexState_WF_);
+        END_LEX_STATE();
+      case LexState_WF_:
+        ACCEPT_LEXEME(Lexeme_WEAK_FAIRNESS);
+        END_LEX_STATE();
       case LexState_PROOF_LEVEL_NUMBER:
         if (iswdigit(lookahead)) {
           proof_step_id_level.push_back(static_cast<char>(lookahead & CHAR_MAX));
@@ -737,6 +761,8 @@ namespace {
     Token_OBVIOUS_KEYWORD,
     Token_OMITTED_KEYWORD,
     Token_QED_KEYWORD,
+    Token_WEAK_FAIRNESS,
+    Token_STRONG_FAIRNESS,
     Token_OTHER
   };
 
@@ -789,6 +815,8 @@ namespace {
       case Lexeme_VARIABLES_KEYWORD: return Token_TERMINATOR;
       case Lexeme_PROOF_STEP_ID: return Token_PROOF_STEP_ID;
       case Lexeme_QED_KEYWORD: return Token_QED_KEYWORD;
+      case Lexeme_STRONG_FAIRNESS: return Token_STRONG_FAIRNESS;
+      case Lexeme_WEAK_FAIRNESS: return Token_WEAK_FAIRNESS;
       case Lexeme_IDENTIFIER: return Token_OTHER;
       case Lexeme_OTHER: return Token_OTHER;
       case Lexeme_END_OF_FILE: return Token_TERMINATOR;
@@ -810,7 +838,7 @@ namespace {
 
     // The starting alignment columnt of the jlist.
     column_index alignment_column;
-
+    
     JunctList() { }
 
     JunctList(JunctType const type, column_index const alignment_column) {
@@ -818,30 +846,24 @@ namespace {
       this->alignment_column = alignment_column;
     }
 
-    unsigned serialize(char* buffer) {
+    unsigned serialize(char* const buffer, bool const is_dry_run) {
       unsigned offset = 0;
-      unsigned byte_count = 0;
       unsigned copied = 0;
 
       // Serialize junction type
       copied = sizeof(uint8_t);
-      buffer[offset] = static_cast<uint8_t>(type);
+      if (!is_dry_run) { buffer[offset] = static_cast<uint8_t>(type); }
       offset += copied;
-      byte_count += copied;
 
       // Serialize alignment column
       copied = sizeof(column_index);
-      memcpy(&buffer[offset], (char*)&alignment_column, copied);
+      if (!is_dry_run) { memcpy(&buffer[offset], (char*)&alignment_column, copied); }
       offset += copied;
-      byte_count += copied;
 
-      return byte_count;
+      return offset;
     }
 
-    unsigned deserialize(const char* const buffer, unsigned const length) {
-      assert(length > 0);
-
-      unsigned byte_count = 0;
+    unsigned deserialize(const char* const buffer) {
       unsigned offset = 0;
       unsigned copied = 0;
 
@@ -849,24 +871,22 @@ namespace {
       copied = sizeof(uint8_t);
       type = JunctType(buffer[offset]);
       offset += copied;
-      byte_count += copied;
 
       // Deserialize alignment column
       copied = sizeof(column_index);
       memcpy((char*)&alignment_column, &buffer[offset], copied);
       offset += copied;
-      byte_count += copied;
-
-      return byte_count;
+      
+      return offset;
     }
   };
   
   /**
-   * A stateful scanner used to parse junction lists.
+   * A stateful scanner used to parse junction lists and proofs.
    */
   struct Scanner {
 
-    //The nested junction lists at the current lexer position.
+    // The nested junction lists at the current lexer position.
     std::vector<JunctList> jlists;
 
     // The nested proofs at the current lexer position.
@@ -884,7 +904,14 @@ namespace {
     Scanner() {
       deserialize(NULL, 0);
     }
-
+    
+    /**
+     * Calculates the serialized size of the scanner.
+     */
+    unsigned serialized_size() {
+      return serialize(NULL, true);
+    }
+    
     /**
      * Serializes the Scanner state into the given buffer.
      *
@@ -892,42 +919,47 @@ namespace {
      * @return Number of bytes written into the buffer.
      */
     unsigned serialize(char* const buffer) {
+      return serialize(buffer, false);
+    }
+
+    /**
+     * Serializes the Scanner state into the given buffer.
+     *
+     * @param buffer The buffer into which to serialize the scanner state.
+     * @param is_dry_run Whether to actually copy the bytes to the buffer.
+     * @return Number of bytes written into the buffer.
+     */
+    unsigned serialize(char* const buffer, const bool is_dry_run) {
       unsigned offset = 0;
-      unsigned byte_count = 0;
       unsigned copied = 0;
 
       const nest_address jlist_depth = static_cast<nest_address>(jlists.size());
       copied = sizeof(nest_address);
-      memcpy(&buffer[offset], &jlist_depth, copied);
+      if (!is_dry_run) { memcpy(&buffer[offset], &jlist_depth, copied); }
       offset += copied;
-      byte_count += copied;
       for (nest_address i = 0; i < jlist_depth; i++) {
-        copied = jlists[i].serialize(&buffer[offset]);
+        char* const buffer_addr = is_dry_run ? NULL : &buffer[offset];
+        copied = jlists[i].serialize(buffer_addr, is_dry_run);
         offset += copied;
-        byte_count += copied;
       }
       
       const nest_address proof_depth = static_cast<nest_address>(proofs.size());
       copied = sizeof(nest_address);
-      memcpy(&buffer[offset], &proof_depth, copied);
+      if (!is_dry_run) { memcpy(&buffer[offset], &proof_depth, copied); }
       offset += copied;
-      byte_count += copied;
       copied = proof_depth * sizeof(proof_level);
-      memcpy(&buffer[offset], proofs.data(), copied);
+      if (!is_dry_run) { memcpy(&buffer[offset], proofs.data(), copied); }
       offset += copied;
-      byte_count += copied;
       
       copied = sizeof(proof_level);
-      memcpy(&buffer[offset], &last_proof_level, copied);
+      if (!is_dry_run) { memcpy(&buffer[offset], &last_proof_level, copied); }
       offset += copied;
-      byte_count += copied;
       
       copied = sizeof(uint8_t);
-      buffer[offset] = static_cast<uint8_t>(have_seen_proof_keyword);
+      if (!is_dry_run) { buffer[offset] = static_cast<uint8_t>(have_seen_proof_keyword); }
       offset += copied;
-      byte_count += copied;
 
-      return byte_count;
+      return offset;
     }
 
     /**
@@ -954,9 +986,10 @@ namespace {
         memcpy(&jlist_depth, &buffer[offset], copied);
         jlists.resize(jlist_depth);
         offset += copied;
+
         for (nest_address i = 0; i < jlist_depth; i++) {
           assert(offset < length);
-          copied = jlists[i].deserialize(&buffer[offset], length - offset);
+          copied = jlists[i].deserialize(&buffer[offset]);
           offset += copied;
         }
       
@@ -965,6 +998,7 @@ namespace {
         memcpy(&proof_depth, &buffer[offset], copied);
         proofs.resize(proof_depth);
         offset += copied;
+
         copied = proof_depth * sizeof(proof_level);
         memcpy(proofs.data(), &buffer[offset], copied);
         offset += copied;
@@ -1238,15 +1272,10 @@ namespace {
      *    improve error reporting since the end-of-module token is missing)
      *
      * @param lexer The tree-sitter lexing control structure.
-     * @param valid_symbols Tokens possibly expected in this spot.
      * @return Whether a jlist-relevant token should be emitted.
      */
-    bool handle_terminator_token(
-      TSLexer* const lexer,
-      const bool* const valid_symbols
-    ) {
-      return is_in_jlist()
-        && emit_dedent(lexer);
+    bool handle_terminator_token(TSLexer* const lexer) {
+      return is_in_jlist() && emit_dedent(lexer);
     }
     
     /**
@@ -1270,7 +1299,6 @@ namespace {
      */
     bool handle_other_token(
       TSLexer* const lexer,
-      const bool* const valid_symbols,
       column_index const next
     ) {
       return is_in_jlist()
@@ -1434,10 +1462,10 @@ namespace {
       } else {
         if (valid_symbols[DEDENT]) {
           // End all jlists before start of proof.
-          return handle_terminator_token(lexer, valid_symbols);
+          return handle_terminator_token(lexer);
         } else {
           // This is a reference to a proof step in an expression.
-          return handle_other_token(lexer, valid_symbols, next);
+          return handle_other_token(lexer, next);
         }
       }
     }
@@ -1461,7 +1489,7 @@ namespace {
         lexer->mark_end(lexer);
         return true;
       } else {
-        return handle_terminator_token(lexer, valid_symbols);
+        return handle_terminator_token(lexer);
       }
     }
     
@@ -1487,7 +1515,7 @@ namespace {
         lexer->mark_end(lexer);
         return true;
       } else {
-        return handle_terminator_token(lexer, valid_symbols);
+        return handle_terminator_token(lexer);
       }
     }
     
@@ -1499,13 +1527,9 @@ namespace {
      * we pop the top proof level off the stack.
      * 
      * @param lexer The tree-sitter lexing control structure.
-     * @param valid_symbols Tokens possibly expected in this spot.
      * @return Whether a token should be emitted.
      */
-    bool handle_qed_keyword_token(
-      TSLexer* const lexer,
-      const bool* const valid_symbols
-    ) {
+    bool handle_qed_keyword_token(TSLexer* const lexer) {
       last_proof_level = get_current_proof_level();
       proofs.pop_back();
       lexer->result_symbol = QED_KEYWORD;
@@ -1513,6 +1537,30 @@ namespace {
       return true;
     }
     
+    /**
+     * Handles the fairness tokens WF_ and SF_.
+     * Need to handle this in an external scanner due to:
+     * https://github.com/tree-sitter/tree-sitter/issues/1615
+     *
+     * @param lexer The tree-sitter lexing control structure.
+     * @param next The column position of the encountered token.
+     * @param keyword_type The specific keyword being handled.
+     * @return Whether a token should be emitted.
+     */
+    bool handle_fairness_keyword_token(
+      TSLexer* const lexer,
+      column_index const next,
+      TokenType keyword_type
+    ) {
+      if (handle_other_token(lexer, next)) {
+        return true;
+      } else {
+        lexer->result_symbol = keyword_type;
+        lexer->mark_end(lexer);
+        return true;
+      }
+    }
+    
     /**
      * Scans for various possible external tokens.
      * 
@@ -1547,7 +1595,7 @@ namespace {
           case Token_COMMENT_START:
             return false;
           case Token_TERMINATOR:
-            return handle_terminator_token(lexer, valid_symbols);
+            return handle_terminator_token(lexer);
           case Token_PROOF_STEP_ID:
             return handle_proof_step_id_token(lexer, valid_symbols, col, proof_step_id_level);
           case Token_PROOF_KEYWORD:
@@ -1559,15 +1607,152 @@ namespace {
           case Token_OMITTED_KEYWORD:
             return handle_terminal_proof_keyword_token(lexer, valid_symbols, OMITTED_KEYWORD);
           case Token_QED_KEYWORD:
-            return handle_qed_keyword_token(lexer, valid_symbols);
+            return handle_qed_keyword_token(lexer);
+          case Token_WEAK_FAIRNESS:
+            return handle_fairness_keyword_token(lexer, col, WEAK_FAIRNESS);
+          case Token_STRONG_FAIRNESS:
+            return handle_fairness_keyword_token(lexer, col, STRONG_FAIRNESS);
           case Token_OTHER:
-            return handle_other_token(lexer, valid_symbols, col);
+            return handle_other_token(lexer, col);
           default:
             return false;
         }
       }
     }
   };
+  
+  /**
+   * A hierarchy of nested stateful scanners.
+   * Each time a PlusCal block is entered, a nested context is created.
+   * Exiting the PlusCal block exits the context.
+   * Multiply-nested PlusCal blocks are supported.
+   */
+  struct NestedScanner {
+
+    // The enclosing context(s) of the PlusCal block.
+    std::vector< std::vector<char> > enclosing_contexts;
+    
+    // The currently-active context.
+    Scanner current_context;
+
+    /**
+     * Initializes a new instance of the NestedScanner object.
+     */
+    NestedScanner() {
+      this->deserialize(NULL, 0);
+    }
+    
+    unsigned serialize(char* const buffer) {
+      unsigned offset = 0;
+      unsigned copied = 0;
+
+      // First write number of enclosing contexts (guaranteed to be >= 1)
+      nest_address const context_depth = this->enclosing_contexts.size() + 1;
+      copied = sizeof(nest_address);
+      memcpy(&buffer[offset], &context_depth, copied);
+      offset += copied;
+      
+      // Then write size of N-1 enclosing contexts
+      for (int i = 0; i < context_depth - 1; i++) {
+        unsigned const context_size = this->enclosing_contexts[i].size();
+        copied = sizeof(unsigned);
+        memcpy(&buffer[offset], &context_size, copied);
+        offset += copied;
+      }
+      
+      // Reserve space for current context size
+      unsigned const current_context_size_offset = offset;
+      copied = sizeof(unsigned);
+      offset += copied;
+      
+      // Serialize N-1 enclosing contexts
+      for (int i = 0; i < this->enclosing_contexts.size(); i++) {
+        std::vector<char>& context = this->enclosing_contexts[i];
+        copied = context.size();
+        memcpy(&buffer[offset], context.data(), copied);
+        offset += copied;
+      }
+
+      // Serialize current context
+      copied = this->current_context.serialize(&buffer[offset]);
+      offset += copied;
+      
+      // Write current context size to reserved position
+      memcpy(&buffer[current_context_size_offset], &copied, sizeof(unsigned));
+      
+      return offset;
+    }
+
+    void deserialize(const char* const buffer, unsigned const length) {
+      this->enclosing_contexts.clear();
+      this->current_context.deserialize(NULL, 0);
+
+      if (length > 0) {
+        unsigned offset = 0;
+        unsigned copied = 0;
+
+        // First item: total number of contexts (guaranteed to be >= 1)
+        nest_address context_depth = 0;
+        copied = sizeof(nest_address);
+        memcpy(&context_depth, &buffer[offset], copied);
+        assert(1 <= context_depth);
+        this->enclosing_contexts.resize(context_depth - 1);
+        offset += copied;
+        
+        // Next N items: size of all contexts
+        std::vector<unsigned> context_sizes;
+        context_sizes.resize(context_depth);
+        copied = context_depth * sizeof(unsigned);
+        memcpy(context_sizes.data(), &buffer[offset], copied);
+        offset += copied;
+        
+        // Deserialize N-1 contexts as enclosing contexts
+        for (int i = 0; i < context_depth - 1; i++) {
+          copied = context_sizes[i];
+          this->enclosing_contexts[i].resize(copied);
+          memcpy(this->enclosing_contexts[i].data(), &buffer[offset], copied);
+          offset += copied;
+        }
+        
+        // Final context is deserialized as current context
+        copied = context_sizes.back();
+        this->current_context.deserialize(&buffer[offset], copied);
+        offset += copied;
+
+        assert(offset == length);
+      }
+    }
+
+    bool scan(TSLexer* const lexer, const bool* const valid_symbols) {
+      // All symbols are marked as valid during error recovery.
+      // We can check for this by looking at the validity of the final
+      // (unused) external symbol, ERROR_SENTINEL.
+      // TODO: actually function during error recovery
+      // https://github.com/tlaplus-community/tree-sitter-tlaplus/issues/19
+      if (valid_symbols[ERROR_SENTINEL]) {
+        return false;
+      } else if (valid_symbols[PCAL_START]) {
+        // Entering PlusCal block; push current context then clear
+        unsigned const expected_size = this->current_context.serialized_size();
+        this->enclosing_contexts.resize(this->enclosing_contexts.size() + 1);
+        this->enclosing_contexts.back().resize(expected_size);
+        unsigned const actual_size = this->current_context.serialize(this->enclosing_contexts.back().data());
+        assert(expected_size == actual_size);
+        this->current_context.deserialize(NULL, 0);
+        lexer->result_symbol = PCAL_START;
+        return true;
+      } else if (valid_symbols[PCAL_END]) {
+        // Exiting PlusCal block; rehydrate context then pop
+        std::vector<char>& next = this->enclosing_contexts.back();
+        this->current_context.deserialize(next.data(), next.size());
+        this->enclosing_contexts.pop_back();
+        lexer->result_symbol = PCAL_END;
+        return true;
+      } else {
+        return current_context.scan(lexer, valid_symbols);
+      }
+    }
+  };
 }
 
 extern "C" {
@@ -1575,13 +1760,13 @@ extern "C" {
   // Called once when language is set on a parser.
   // Allocates memory for storing scanner state.
   void* tree_sitter_tlaplus_external_scanner_create() {
-    return new Scanner();
+    return new NestedScanner();
   }
 
   // Called once parser is deleted or different language set.
   // Frees memory storing scanner state.
   void tree_sitter_tlaplus_external_scanner_destroy(void* const payload) {
-    Scanner* const scanner = static_cast<Scanner*>(payload);
+    NestedScanner* const scanner = static_cast<NestedScanner*>(payload);
     delete scanner;
   }
 
@@ -1591,7 +1776,7 @@ extern "C" {
     void* const payload,
     char* const buffer
   ) {
-    Scanner* scanner = static_cast<Scanner*>(payload);
+    NestedScanner* scanner = static_cast<NestedScanner*>(payload);
     return scanner->serialize(buffer);
   }
 
@@ -1602,7 +1787,7 @@ extern "C" {
     const char* const buffer,
     unsigned const length
   ) {
-    Scanner* const scanner = static_cast<Scanner*>(payload);
+    NestedScanner* const scanner = static_cast<NestedScanner*>(payload);
     scanner->deserialize(buffer, length);
   }
 
@@ -1612,7 +1797,7 @@ extern "C" {
     TSLexer* const lexer,
     const bool* const valid_symbols
   ) {
-    Scanner* const scanner = static_cast<Scanner*>(payload);
+    NestedScanner* const scanner = static_cast<NestedScanner*>(payload);
     return scanner->scan(lexer, valid_symbols);
   }
 }
